#!/usr/bin/perl

# Author^ Vasyl Ostrovskyi <vasyusya@gmail.com>
# Version: 1.00 beta
# License: GPL
# Description: Splits roads in Polish format maps into parts at routing nodes
# Usage: ./mpsplit.pl filename.mp
# output in _out.mp


# You can change these values
my $logfile="_logfile.txt";
my $outfile="_out.mp";

# No customizable parts below
use strict;
use Geo::Parse::PolishFormat;
use Data::Dumper;

my $mpfile=$ARGV[0];
die "Usage: perl _mpsplit.pl <filename>.mp\n" unless $mpfile;

my %restricts;
my %linkid;
my $lastlink=0;
my $lastnode=0;
my $lastroadid=0;
my %cities;
my %regions;
my %countries;
my @zipcodes;
my %types;
my $counter=0;
my %splitpoints;
my %zlevel;
my %fromroads;
my %toroads;
my %addressdata;
my $insertpoint;
my %maxsize;
my $points;
my $poi;
my $lines;
my $roads;
my $areas;


# what we do:
# read restricts into hash
# if the feature is not a polyline or restrict print verbosely
# if the feature is a polyline without NodNN do the same
# if the feature is a polyline with NodNN
#   split into chunks at each Nod
#   for each chunk print verbosely all data except Data0 RoadID NodNN NumbersNN
#                  modify RoadId and print new value
#                  print the corresponding part of Data0
#                  check, recalculate and print new NumbersNN
#                  recalculate and write HLevelNN
#                  fix restricts
# after all done print modified restricts



# Create log
open (LOG, ">$logfile") or die;
open (OUT, ">$outfile") or die;
# Read types 

my $p = Geo::Parse::PolishFormat -> new ();

$counter=0;
print "Reading turn restrictions...\n";
$p -> parse ($mpfile, \&readrestricts);

warn "last node $lastnode, last road $lastroadid\n";

$counter=0;
print "\nSplitting...\n";
$p -> parse ($mpfile, \&splitmp);

#$counter=0;
print "\nUpdating restricts...\n";
#$p -> parse ($mpfile, \&restrict);
restrict();	

print "\nClosing files...\n";

close LOG;

print "Done!\n";







# Reads turn restrictions and fills the array. 
# $restricts{$nodeid}{'nodes'} lists all restrictions
# at the node $nodeid in the form from,in,to nodeids
# $restricts{$nodeid}{'roads'} lists similarly pairs of from,to roads
#
# also we calculate $lastnode and $lastroadid

sub readrestricts{
  print "$counter\r";
  $counter++;
  if ($_[0]{'name'} eq 'Restrict'){
#    print "r";
    $counter++;
    my $roads= $_[0]{'attributes'}{'TraffRoads'};
    my $nodes= $_[0]{'attributes'}{'TraffPoints'};
    my $nodeid=$_[0]{'attributes'}{'Nod'}[0];
    my $time= $_[0]{'attributes'}{'Time'};
    push @{$restricts{$nodeid}{'nodes'}}, $nodes;
    push @{$restricts{$nodeid}{'roads'}}, $roads;
    push @{$restricts{$nodeid}{'time'}}, $time;
  };
  if ( defined $_[0]{'attributes'}{'RoadID'}){
    $lastroadid = $_[0]{'attributes'}{'RoadID'} if ($_[0]{'attributes'}{'RoadID'}>$lastroadid);
  }
  if ( defined $_[0]{'collections'}{'Nod'}){
    my @nodelist=@{$_[0]{'collections'}{'Nod'}};
    shift @nodelist;
    foreach my $nod (@nodelist){
    $lastnode=@{$nod}[1] if ($lastnode<@{$nod}[1]);
    }
  }
#  warn Dumper %restricts;
};




# Reads data and writes shapes and metadata
sub splitmp{
  $counter++;
  print "$counter\r";
  return if ($_[0]{'name'} eq 'Restrict');
  if (defined $_[0]{'collections'}{'Nod'}){
      # this is a road 
      #warn Dumper $_[0]{'collections'};
      my $roadid=$_[0]{'attributes'}{'RoadID'} or die "Road without RoadId!";
      # Get vertices and routing nodes
      my @coords=@{$_[0]{'collections'}{'Data'}[0]};
      my @nodelist=@{$_[0]{'collections'}{'Nod'}};
      #warn Dumper @nodelist;
      #remove empty
      shift @nodelist;
      #warn Dumper @nodelist;
      # split into links      
           
      my $nod;
      # Split into chains
      if ($nodelist[0][0] ne '0'){
        warn "\nFirst node must be routable, RoadId=$roadid! Fixed.";
        print LOG "\nFirst node must be routable, RoadId=$roadid! Fixed.\n";
	$lastnode++;
	unshift @nodelist,[ (0,$lastnode,0) ];
	#warn Dumper @nodelist;
      }
      if ($nodelist[$#nodelist][0] ne $#coords){
        warn "\nLast node must be routable, RoadId=$roadid! Fixed.";
        print LOG "\nLast node must be routable, RoadId=$roadid! Fixed.\n";
	$lastnode++;
	push @nodelist,[ ($#coords,$lastnode,0) ];
	#warn Dumper @nodelist;
      }
      #die unless defined $nodelist[1];
      if (defined $nodelist[1]){
        for $nod (1 .. $#nodelist){
  	  my @modified = ('RoadID','Data','Nod','Numbers','HLevel');
          print OUT "\n[".$_[0]{'name'}."]\n";
	  foreach my $l (@{$_[0]{'lines'}}){
	    print OUT $l."\n" unless (grep $l=~/^$_/,@modified);
	  }    
	  $lastroadid++;
	  my $newroadid=$lastroadid;
	  print OUT "RoadID=$newroadid\n";
	  my $headnod=$nodelist[$nod][1];
	  my $tailnod=$nodelist[$nod-1][1];
	  my $headpoint=$nodelist[$nod][0];
	  my $tailpoint=$nodelist[$nod-1][0];
	  my @ccc=@coords[$tailpoint..$headpoint];
	  my @str;
	  undef @str;
	  foreach my $c (@ccc){
	    push @str,"(".join (",", (@{$c})).")";
	  }
	  print OUT "Data0=".join(",", (@str))."\n";
	  print OUT "Nod1=0,".$nodelist[$nod-1][1].",".$nodelist[$nod-1][2]."\n";
	  print OUT "Nod2=".$#ccc.",".$nodelist[$nod][1].",".$nodelist[$nod][2]."\n";
          # Numbers
	  if (defined $_[0]{'collections'}{Numbers}){
	    #print OUT ";Numbers\n";
	    my @numbers= @{$_[0]{'collections'}{Numbers}};
	    shift @numbers;
	    my $prevn;
	    my @ndata;
	    undef $prevn;
	    #warn Dumper @numbers;
	    my $i=1;
	    foreach my $ninfo (@numbers){
	      my ($n)=split ',', $ninfo;
	      if (defined $prevn){
	        if ($prevn<$tailpoint){
		  ($prevn,@ndata)=split ',', $ninfo;
		  next;
		}
		last if ($n>$headpoint);
		print OUT "Numbers$i=".($prevn-$tailpoint).",".join(',',@ndata)."\n";
		$i++;		
	      }
	      ($prevn,@ndata)=split ',', $ninfo;
	      #warn Dumper @ndata;
	    }
	    unless (($prevn < $tailpoint) || ($prevn > $headpoint)){
	      warn "Numbering skewed" if ($headpoint < $#ccc);
              print OUT "Numbers$i=";
	      print OUT $prevn-$tailpoint . ",".join(',',@ndata)."\n";
	    }
	  }
	  # HLevel
	  if (defined $_[0]{'collections'}{'HLevel'}[0]){
 	    my $zz = $_[0]{'collections'}{'HLevel'}[0];
	    $zz =~ s/\((.*)\)/$1/;
	    my @zzz=split('\),\(',$zz);
	    # @zzz contains strings 'vertex,zlevel'
	    #warn Dumper @zzz;
	    my @znew=();
	    foreach my $i (@zzz){
	      my ($pointer,$zlevel)=split ',',$i;
	      next if ($pointer<$tailpoint);
	      last if ($pointer>$headpoint);
	      my $tmpz="(".($pointer-$tailpoint).",".$zlevel.")";
	      #warn $tmpz;
	      push @znew, $tmpz;
	    }
	    if ($#znew){
	      print OUT "HLevel0=".join (",", @znew)."\n";
	    }
	  }
	  print OUT "[END]\n"; 
	  # fix routing
	  # at tail
	  if (defined $restricts{$tailnod}){
	    for my $k (0 .. $#{$restricts{$tailnod}{'roads'}}){
	      my ($fromroad,$toroad)=split ",",$restricts{$tailnod}{'roads'}[$k];
	      my $fromnod=$restricts{$tailnod}{'nodes'}[$k];
              $fromnod =~ s/^(\d+),\d+,\d+$/$1/;
	      my $tonod=$restricts{$tailnod}{'nodes'}[$k];
              $tonod =~ s/^\d+,\d+,(\d+)$/$1/;
	      if ($roadid eq $fromroad){
	        if ($fromnod eq $headnod){
		  #warn "going to modify $k-th restrict at $tailnod to $newroadid,$toroad";
	          #$restricts{$tailnod}{'roads'}[$k] = $newroadid.",".$toroad;
	          $restricts{$tailnod}{'roads'}[$k] =~s/^\d+,(\d+)$/$newroadid,$1/;
	        }
	      }
	      if ($roadid eq $toroad){
	        if ($tonod eq $headnod){
	          #$restricts{$tailnod}{'roads'}[$k] = $fromroad.",".$newroadid;
	          $restricts{$tailnod}{'roads'}[$k] =~s/^(\d+),\d+$/$1,$newroadid/;
	        }
	      }
	    }
          }
	  # at head
	  if (defined $restricts{$headnod}){
	    for my $k (0 .. $#{$restricts{$headnod}{'roads'}}){
	      my ($fromroad,$toroad)=split ",",$restricts{$headnod}{'roads'}[$k];
	      my $fromnod=$restricts{$headnod}{'nodes'}[$k];
	      $fromnod =~ s/^(\d+),\d+,\d+$/$1/;
	      my $tonod=$restricts{$headnod}{'nodes'}[$k];
	      $tonod =~ s/^\d+,\d+,(\d+)$/$1/;
	      if ($roadid eq $fromroad){
	        if ($fromnod eq $tailnod){
	          #$restricts{$headnod}{'roads'}[$k] = $newroadid.",".$toroad;
	          $restricts{$headnod}{'roads'}[$k] =~s/^\d+,(\d+)$/$newroadid,$1/;
	        }
	      }
	      if ($roadid eq $toroad){
	        if ($tonod eq $tailnod){
	          #$restricts{$headnod}{'roads'}[$k] = $fromroad.",".$newroadid;
	          $restricts{$headnod}{'roads'}[$k] =~s/^(\d+),\d+$/$1,$newroadid/;
	        }
	      }
	    }
	  }
        }
      }
      else{
        warn "Problem with routing nodes for road $roadid! Need 2 or more nodes";
	print LOG "Problem with routing nodes for road $roadid! Need 2 or more nodes\n";
      }
      # if the first node in not routable (this should not happen)
  }
  else{
    print OUT "\n[".$_[0]{'name'}."]\n";
    foreach my $l (@{$_[0]{'lines'}}){
      print OUT $l."\n";
    }    
    my @features = ('POI','POLYGON','POLYLINE');
    unless (grep $_ eq $_[0]{'name'},@features){
      print OUT "[END-".$_[0]{'name'}."]\n";
    }
    else{
      print OUT "[END]\n"; 
    }
    #warn Dumper    $_[0]{'attributes'};
    if (defined $_[0]{'attributes'}{'RoadID'}){
      warn "Road ".$_[0]{'attributes'}{'RoadID'}." without routing nodes!";
      print LOG "Road ".$_[0]{'attributes'}{'RoadID'}." without routing nodes!\n";
    }
  }
};


sub restrict{
  my @nodes=keys %restricts;
  foreach my $k (@nodes){
    for my $j (0..$#{$restricts{$k}{'nodes'}}){
      print OUT "\n[Restrict]\n";
      print OUT "Nod=".$k."\n";
      print OUT "TraffPoints=".$restricts{$k}{'nodes'}[$j]."\n";
      print OUT "TraffRoads=".$restricts{$k}{'roads'}[$j]."\n";
      print OUT "Time=".$restricts{$k}{'time'}[$j]."\n";
      print OUT "[END-Restrict]\n"
    }
  }
}

